/**************************************************************************

	ns

	func_code.c

	Copyright (C) 1996-2002 by Tomoaki Nakashima. All rights reserved.
		http://www.nakka.com/
		nakka@nakka.com

**************************************************************************/

/**************************************************************************
	Include Files
**************************************************************************/

#include <windows.h>

#include "script.h"
#include "script_memory.h"
#include "script_string.h"
#include "script_utility.h"


/**************************************************************************
	Define
**************************************************************************/

#define ESC						0x1B

#define	IsKanji(i)				((i >= 0x81 && i <= 0x9F) || (i >= 0xE0 && i <= 0xFC))
#define	IsHankaku(i)			(i >= 0xa0 && i <= 0xdf)

#ifdef UNICODE
#define tstrcpy					strcpy
#else
#define tstrcpy					lstrcpy
#endif


/**************************************************************************
	Global Variables
**************************************************************************/

extern char AppDir[];


/**************************************************************************
	Local Function Prototypes
**************************************************************************/

static BOOL IsLeadChar(const char c);
static int GetKanjiType(unsigned char *str);
static int HanToZen(unsigned int *zenkaku, unsigned char *str);
static void JisShift(int *ph,int *pl);
static char *SjisToJis(unsigned char *buf,unsigned char *ret);
static void SjisShift(int *ph, int *pl);
static char *JisToSjis(char *buf, char *ret);
static char *SjisToEuc(char *buf, char *ret);
static char *EucToSjis(char *buf, char *ret);
static void ConvStrCode(char *buf, char *ret);
static char *dBase(char *buf, char *ret);
static void eBase(char *buf, char *ret, int size);


/******************************************************************************

	IsLeadChar

	Sp1oCgڂǂ`FbN

******************************************************************************/

static BOOL IsLeadChar(const char c)
{
	return IsDBCSLeadByte((BYTE)c);
}


/******************************************************************************

	GetKanjiType

	R[h̎ނ擾

	0 - ASCII
	1 - SJISpJi
	2 - SJIS
	4 - EUCpJi
	8 - EUC
	9 - EUCSJISpJi
	16 - VJIS
	32 - JIS

******************************************************************************/

static int GetKanjiType(unsigned char *str)
{
    int val = 0;
    unsigned char b1, b2, b3;

    b1 = *str++;
    b2 = *str++;
    b3 = *str;
    if(b1 == 0x1b) {
        if (b2 == '$' && b3 == 'B') return 16;
        if (b2 == '$' && b3 == '@') return 32;
        return 0;
    }
    if( b1 >= 0xa0 && b1 <= 0xdf) val |= 1;
    if((b1 >= 0x81 && b1 <= 0x9f ||
         b1 >= 0xe0 && b1 <= 0xfc) &&
        (b2 >= 0x40 && b2 <= 0xfc && b2 != 0x7f)) val |= 2;
    if(b1 == 0x8e && (b2 >= 0xa0 && b2 <= 0xdf)) val |= 4;
    if((b1 >= 0xa1 && b1 <= 0xfe) &&
        (b2 >= 0xa1 && b1 <= 0xfe)) val |= 8;

    return val;
}


/******************************************************************************

	HanToZen

	pJiSpJiɕϊ

******************************************************************************/

static int HanToZen(unsigned int *zenkaku, unsigned char *str)
{
	const unsigned int z[64] = {
		0x2121,0x2123,0x2156,0x2157,0x2122,0x2126,0x2572,0x2521,
		0x2523,0x2525,0x2527,0x2529,0x2563,0x2565,0x2567,0x2543,
		0x213c,0x2522,0x2524,0x2526,0x2528,0x252a,0x252b,0x252d,
		0x252f,0x2531,0x2533,0x2535,0x2537,0x2539,0x253b,0x253d,
		0x253f,0x2541,0x2544,0x2546,0x2548,0x254a,0x254b,0x254c,
		0x254d,0x254e,0x254f,0x2552,0x2555,0x2558,0x255b,0x255e,
		0x255f,0x2560,0x2561,0x2562,0x2564,0x2566,0x2568,0x2569,
		0x256a,0x256b,0x256c,0x256d,0x256f,0x2573,0x212b,0x212c};
	typedef struct {
		unsigned char han;
		unsigned int zen;
	} TBL;
	const TBL daku[] = {
		{0xb3,0x2574},{0xb6,0x252c},{0xb7,0x252e},{0xb8,0x2530},
		{0xb9,0x2532},{0xba,0x2534},{0xbb,0x2536},{0xbc,0x2538},
		{0xbd,0x253a},{0xbe,0x253c},{0xbf,0x253e},{0xc0,0x2540},
		{0xc1,0x2542},{0xc2,0x2545},{0xc3,0x2547},{0xc4,0x2549},
		{0xca,0x2550},{0xcb,0x2553},{0xcc,0x2556},{0xcd,0x2559},
		{0xce,0x255c},{0,0}};
	const TBL handaku[] = {
		{0xca,0x2551},{0xcb,0x2554},{0xcc,0x2557},{0xcd,0x255a},
		{0xce,0x255d},{0,0}};
	int i;

	if(*(str+1) == 0xde){           /*  */
		for(i = 0;(daku + i)->zen != 0;i++)
			if(*str == (daku + i)->han){
			*zenkaku = (daku + i)->zen;
			return 2;
		}
	}else if(*(str+1) == 0xdf){    /*  */
	for(i = 0;(handaku + i)->zen != 0;i++)
		if(*str == (handaku + i)->han){
			*zenkaku = (handaku + i)->zen;
			return 2;
		}
	}
	*zenkaku = z[*str - 0xa0];
	return 1;
}


/******************************************************************************

	JisShift

	SJIS2oCgJISɕϊ

******************************************************************************/

static void JisShift(int *ph,int *pl)
{
	*ph = *ph & 0xff;
	*pl = *pl & 0xff;

	if(*ph <= 0x9F){
		if(*pl < 0x9F){
			*ph = (*ph << 1) - 0xE1;
		}else{
			*ph = (*ph << 1) - 0xE0;
		}
	}else{
		if(*pl < 0x9F){
			*ph = (*ph << 1) - 0x161;
		}else{
			*ph = (*ph << 1) - 0x160;
		}
	}

	if(*pl < 0x7F){
		*pl -= 0x1F;
	}else if(*pl < 0x9F){
		*pl -= 0x20;
	}else{
		*pl -= 0x7E;
	}
}


/******************************************************************************

	SjisToJis

	SJISJISɕϊ

******************************************************************************/

static char *SjisToJis(unsigned char *buf,unsigned char *ret)
{
	unsigned char *p,*r;
	int knaji;
	int c,d;
	int rc;
	unsigned int e;
	unsigned int hib, lob;

	p = buf;
	r = ret;
	knaji = 0;

	while((c = *(p++)) != '\0'){
		switch(knaji)
		{
		case 0:
			if(IsKanji((unsigned char)c) == TRUE || IsHankaku((unsigned char)c) == TRUE){
				*(r++) = ESC;
				*(r++) = '$';
				*(r++) = 'B';

				if(IsKanji((unsigned char)c) == TRUE){
					d = *(p++);
					JisShift(&c,&d);
					*(r++) = c;
					*(r++) = d;
				}else{
					rc = HanToZen(&e,p - 1);
					if(rc != 1){
						p++;
					}
					hib = (e >> 8) & 0xff;
					lob = e & 0xff;
					*(r++) = hib;
					*(r++) = lob;
				}
				knaji = 1;
			}else{
				*(r++) = c;
			}
			break;
		case 1:
			if(IsKanji((unsigned char)c) == FALSE && IsHankaku((unsigned char)c) == FALSE){
				*(r++) = ESC;
				*(r++) = '(';
				*(r++) = 'B';
				*(r++) = c;
				knaji = 0;
			}else{
				if(IsKanji((unsigned char)c) == TRUE){
					d = *(p++);
					JisShift(&c,&d);
					*(r++) = c;
					*(r++) = d;
				}else{
					rc = HanToZen(&e,p - 1);
					if(rc != 1){
						p++;
					}
					hib = (e >> 8) & 0xff;
					lob = e & 0xff;
					*(r++) = hib;
					*(r++) = lob;
				}
			}
			break;
		}
	}
	if(knaji == 1){
		*(r++) = ESC;
		*(r++) = '(';
		*(r++) = 'B';
	}
	*r = '\0';
	return r;
}


/******************************************************************************

	SjisShift

	JIS2oCgSJISɕϊ

******************************************************************************/

static void SjisShift(int *ph, int *pl)
{
	if(*ph & 1){
		if(*pl < 0x60){
			*pl+=0x1F;
		}else{
			*pl+=0x20;
		}
	}else{
		*pl+=0x7E;
	}
	if(*ph < 0x5F){
		*ph = (*ph + 0xE1) >> 1;
	}else{
		*ph = (*ph + 0x161) >> 1;
	}
}


/******************************************************************************

	JisToSjis

	JISSJISɕϊ

******************************************************************************/

static char *JisToSjis(char *buf, char *ret)
{
	unsigned int c, d, j;
	BOOL jiskanji, hankaku;
	char *p, *r;

	*ret = '\0';

	p = buf;
	r = ret;
	j = 0;
	jiskanji = FALSE;
	hankaku = FALSE;
	while(*p != '\0'){
		j++;
		c = *(p++);
		if(c == ESC){
			if(*p == '\0'){
				break;
			}
			if((c = *(p++)) == '$'){
				if((c = *(p++)) == '@' || c == 'B'){
					jiskanji = TRUE;
				}else{
					*(r++) = ESC;
					*(r++) = '$';
					if(c != '\0'){
						*(r++) = c;
					}
				}
			}else if(c == '('){
				if((c = *(p++)) == 'H' || c == 'J' || c == 'B'){
					jiskanji = FALSE;
				}else{
					jiskanji = FALSE;
					*(r++) = ESC;
					*(r++) = '(';
					if(c != '\0'){
						*(r++) = c;
					}
				}
			}else if(c == '*'){
				if((c = *(p++)) == 'B'){
					hankaku = FALSE;
				}else if(c == 'I'){
					hankaku = TRUE;
				}
			}else if(hankaku == TRUE && c == 'N'){
				c = *(p++);
				*(r++) = c + 0x80;
			}else if(c == 'K'){
				jiskanji = TRUE;
			}else if(c == 'H'){
				jiskanji = FALSE;
			}else{
				*(r++) = ESC;
				if(c != '\0'){
					*(r++) = c;
				}
			}
		}else if(jiskanji && c >= 0x21 && c <= 0x7E){
			if(*p == '\0'){
				break;
			}
			if((d = *(p++)) >= 0x21 && d <= 0x7E){
				SjisShift((int *)&c, (int *)&d);
			}
			*(r++) = c;
			if(d != '\0'){
				*(r++) = d;
			}
		}else if(c >= 0xA1 && c <= 0xFE){
			if(*p == '\0'){
				break;
			}
			if((d = *(p++)) >= 0xA1 && d <= 0xFE){
				d &= 0x7E;
				c &= 0x7E;
				SjisShift((int *)&c, (int *)&d);
			}
			*(r++) = c;
			if(d != '\0'){
				*(r++) = d;
			}
		}else if(c == 0x0E){
			while(*p != '\0' && *p != 0x0F){
				*(r++) = *(p++) + 0x80;
			}
			if(*p == 0x0F){
				p++;
			}
		}else{
			*(r++) = c;
		}
	}
	*r = '\0';
	return r;
}


/******************************************************************************

	SjisToEuc

	SJISEUCɕϊ

******************************************************************************/

static char *SjisToEuc(char *buf, char *ret)
{
	char *p, *r;
	unsigned int c, d;
    
	p = buf;
	r = ret;
	while(*p != '\0'){
		c = *(p++) & 0xFF;
		if(*p == '\0'){
			break;
		}
		d = *p & 0xFF;
		if(IsKanji((unsigned char)c) == TRUE){
			c -= (c <= 0x9F) ? 0x71 : 0xB1;
			c = (c << 1) + 1;
			if(d >= 0x9E){
				d -= 0x7E;
				c++;
			}else if(d > 0x7F){
				d -= 0x20;
			}else{
				d -= 0x1F;
			}
			c |= 0x80;
			d |= 0x80;
			*(r++) = (char)c;
			*(r++) = (char)d;
			p++;
		}else{
			*(r++) = (char)c;
		}
	}
	*r = '\0';
	return r;
}


/******************************************************************************

	EucToSjis

	EUCSJISɕϊ

******************************************************************************/

static char *EucToSjis(char *buf, char *ret)
{
	char *p, *r;
	unsigned int c, d;

	p = buf;
	r = ret;
	while(*p != '\0'){
		c = *(p++) & 0xFF;
		if(*p == '\0'){
			break;
		}
		d = *p & 0xFF;
		if((c >= 0xa1 && c <= 0xfe) && (d >= 0xa1 && d <= 0xfe)){
			//EUCJISɕϊ
			c &= ~0x80;
			d &= ~0x80;

			//JISSJISɕϊ
			SjisShift((int *)&c, (int *)&d);

			*(r++) = (char)c;
			*(r++) = (char)d;
			p++;
		}else{
			*(r++) = (char)c;
		}
	}
	*r = '\0';
	return r;
}


/******************************************************************************

	ConvStrCode

	JISEUC̊R[hSJISɕϊ

******************************************************************************/

static void ConvStrCode(char *buf, char *ret)
{
	char *p;
	int rc;
	BOOL eucflag = FALSE;

	p = buf;
	while(*p != '\0'){
		rc = GetKanjiType(p);
		switch(rc)
		{
		case 0:		//ASCII
			p++;
			break;

		case 2:		//SJIS
#ifdef UNICODE
			strcpy(ret, buf);
#else
			lstrcpy(ret, buf);
#endif
			return;

		case 4:		//EUCpJi
		case 8:		//EUC
			EucToSjis(buf, ret);
			return;

		case 9:		//EUCSJISpJi
			eucflag = TRUE;
			p += 2;
			break;

		case 16:	//VJIS
		case 32:	//JIS
			JisToSjis(buf, ret);
			return;

		default:
			p += 2;
			break;
		}
	}
	if(eucflag == TRUE){
		EucToSjis(buf, ret);
		return;
	}
#ifdef UNICODE
	strcpy(ret, buf);
#else
	lstrcpy(ret, buf);
#endif
}


/******************************************************************************

	dBase

	BASE64̃fR[hs

******************************************************************************/

static char *dBase(char *buf, char *ret)
{
	int b, c, d, i;
	int *Base;
	unsigned char bb, *bf, *rf;

	if(*buf == '\0'){
		*ret = '\0';
		return ret;
	}

	Base = (int *)LocalAlloc(LPTR, sizeof(int) * 256);
	if(Base == NULL){
		tstrcpy(ret, buf);
		return ret;
	}
	for(i = 0;i < 26;i++){
		*(Base + 'A' + i) = i;
		*(Base + 'a' + i) = i + 26;
	}
	for(i = 0;i < 10;i++){
		*(Base + '0' + i) = i + 52;
	}
	*(Base + '+') = 62;
	*(Base + '/') = 63;

	bf = buf;
	rf = ret;

	for(;*bf != '\0' &&
		(*bf == ' ' || *bf == '\t' || *bf == '\r' || *bf == '\n');bf++);

	while(1){
		if((bb = *bf) == '=' || bb == '\0'){
			break;
		}
		c = *(Base + bb);

		bf++;
		for(;*bf != '\0' &&
			(*bf == ' ' || *bf == '\t' || *bf == '\r' || *bf == '\n');bf++);
		if((bb = *bf) == '=' || bb == '\0'){
			break;
		}
        d = *(Base + bb);
        b = c & 0x3;
        *(rf++) = (c << 2) | (d >> 4);

		bf++;
		for(;*bf != '\0' &&
			(*bf == ' ' || *bf == '\t' || *bf == '\r' || *bf == '\n');bf++);
		if((bb = *bf) == '=' || bb == '\0'){
			break;
		}
		c = *(Base + bb);
        b = d & 0xF;
        *(rf++) = (b << 4) | (c >> 2);

		bf++;
		for(;*bf != '\0' &&
			(*bf == ' ' || *bf == '\t' || *bf == '\r' || *bf == '\n');bf++);
		if((bb = *bf) == '=' || bb == '\0'){
			break;
		}
		d = *(Base + bb);
		bf++;
		for(;*bf != '\0' &&
			(*bf == ' ' || *bf == '\t' || *bf == '\r' || *bf == '\n');bf++);

        b = c & 0x3;
        *(rf++) = (b << 6) | d;
	}
	*rf = '\0';

	LocalFree(Base);
	return rf;
}


/******************************************************************************

	eBase

	BASE64̃GR[hs

******************************************************************************/

static void eBase(char *buf, char *ret, int size)
{
	char tmp, tmp2;
	char *r;
	int c, i;
	const char Base[] =
		"ABCDEFGHIJKLMNOPQRSTUVWXYZ"
		"abcdefghijklmnopqrstuvwxyz"
		"0123456789+/";

	i = 0;
	r = ret;
	while(1){
		if((size == 0 && *(buf + i) == '\0') || (size > 0 && size <= i)){
			break;
		}
		c = (*(buf + i) & 0xFC) >> 2;
		*(r++) = *(Base + c);
		i++;

		if((size == 0 && *(buf + i) == '\0') || (size > 0 && size <= i)){
			*(buf + i) = 0;
			tmp2 = (char)(*(buf + i - 1) << 4) & 0x30;
			tmp = (char)(*(buf + i) >> 4) & 0xF;
			c = tmp2 | tmp;
			*(r++) = *(Base + c);
			*(r++) = '=';
			*(r++) = '=';
			break;
		}
		tmp2 = (char)(*(buf + i - 1) << 4) & 0x30;
		tmp = (char)(*(buf + i) >> 4) & 0xF;
		c = tmp2 | tmp;
		*(r++) = *(Base + c);

		if((size == 0 && *(buf + i + 1) == '\0') || (size > 0 && size <= (i + 1))){
			*(buf + i + 1) = 0;
			tmp2 = (char)(*(buf + i) << 2) & 0x3C;
			tmp = (char)(*(buf + i + 1) >> 6) & 0x3;
			c = tmp2 | tmp;
			*(r++) = *(Base + c);
			*(r++) = '=';
			break;
		}

		tmp2 = (char)(*(buf + i) << 2) & 0x3C;
		tmp = (char)(*(buf + i + 1) >> 6) & 0x3;
		c = tmp2 | tmp;
		*(r++) = *(Base + c);
		i++;

		c = *(buf + i) & 0x3F;
		*(r++) = *(Base + c);
		i++;
	}
	*r = '\0';
}


/******************************************************************************

	_sfunc_isleadchar

	Sp1oCgڂǂ`FbN

******************************************************************************/

int SFUNC _sfunc_isleadchar(EXECINFO *ei, VALUEINFO *param, VALUEINFO *ret, char *ErrStr)
{
	if(param == NULL){
		return -2;
	}

	if(param->v->Type == TYPE_STRING && IsLeadChar(*param->v->u.Value) == TRUE){
		ret->v->u.iValue = 1;
	}else{
		ret->v->u.iValue = 0;
	}
	ret->v->Type = TYPE_INTEGER;
	return 0;
}


/******************************************************************************

	_sfunc_getcodetype

	R[h̎ނ擾

******************************************************************************/

int SFUNC _sfunc_getcodetype(EXECINFO *ei, VALUEINFO *param, VALUEINFO *ret, char *ErrStr)
{
	char *p;
	int rc;
	BOOL eucflag = FALSE;

	if(param == NULL){
		return -2;
	}

	if(param->v->Type == TYPE_STRING){
		p = param->v->u.Value;
		ret->v->u.iValue = 0;
		while(*p != '\0'){
			rc = GetKanjiType(p);
			switch(rc)
			{
			case 0:		//ASCII
				p++;
				break;

			case 9:		//EUCSJISpJi
				eucflag = TRUE;
				p += 2;
				break;

			case 2:		//SJIS
				ret->v->u.iValue = 1;
				break;

			case 4:		//EUCpJi
			case 8:		//EUC
				ret->v->u.iValue = 3;
				break;

			case 16:	//VJIS
			case 32:	//JIS
				ret->v->u.iValue = 2;
				break;

			default:
				p += 2;
				break;
			}
			if(ret->v->u.iValue != 0){
				break;
			}
		}
		if(eucflag == TRUE && ret->v->u.iValue == 0){
			ret->v->u.iValue = 3;
		}
	}else{
		ret->v->u.iValue = 0;
	}
	ret->v->Type = TYPE_INTEGER;
	return 0;
}


/******************************************************************************

	_sfunc_sjistojis

	SJISJISɕϊ

******************************************************************************/

int SFUNC _sfunc_sjistojis(EXECINFO *ei, VALUEINFO *param, VALUEINFO *ret, char *ErrStr)
{
	char *str;

	if(param == NULL){
		return -2;
	}

	str = VariableToString(param);

	ret->v->u.Value = mem_alloc(sizeof(char) * (tstrlen(str) * 4 + 1));
	if(ret->v->u.Value == NULL){
		tstrcpy(ErrStr, "SjisToJis: "ERR_ALLOC);
		mem_free(&str);
		return -1;
	}
	SjisToJis(str, ret->v->u.Value);
	ret->v->Type = TYPE_STRING;
	mem_free(&str);
	return 0;
}


/******************************************************************************

	_sfunc_jistosjis

	JISSJISɕϊ

******************************************************************************/

int SFUNC _sfunc_jistosjis(EXECINFO *ei, VALUEINFO *param, VALUEINFO *ret, char *ErrStr)
{
	char *str;

	if(param == NULL){
		return -2;
	}

	str = VariableToString(param);

	ret->v->u.Value = mem_alloc(sizeof(char) * (tstrlen(str) + 1));
	if(ret->v->u.Value == NULL){
		tstrcpy(ErrStr, "JisToSjis: "ERR_ALLOC);
		mem_free(&str);
		return -1;
	}
	JisToSjis(str, ret->v->u.Value);
	ret->v->Type = TYPE_STRING;
	mem_free(&str);
	return 0;
}


/******************************************************************************

	_sfunc_sjistoeuc

	SJISEUCɕϊ

******************************************************************************/

int SFUNC _sfunc_sjistoeuc(EXECINFO *ei, VALUEINFO *param, VALUEINFO *ret, char *ErrStr)
{
	char *str;

	if(param == NULL){
		return -2;
	}

	str = VariableToString(param);

	ret->v->u.Value = mem_alloc(sizeof(char) * (tstrlen(str) + 1));
	if(ret->v->u.Value == NULL){
		tstrcpy(ErrStr, "EucToSjis: "ERR_ALLOC);
		mem_free(&str);
		return -1;
	}
	SjisToEuc(str, ret->v->u.Value);
	ret->v->Type = TYPE_STRING;
	mem_free(&str);
	return 0;
}



/******************************************************************************

	_sfunc_euctosjis

	EUCSJISɕϊ

******************************************************************************/

int SFUNC _sfunc_euctosjis(EXECINFO *ei, VALUEINFO *param, VALUEINFO *ret, char *ErrStr)
{
	char *str;

	if(param == NULL){
		return -2;
	}

	str = VariableToString(param);

	ret->v->u.Value = mem_alloc(sizeof(char) * (tstrlen(str) + 1));
	if(ret->v->u.Value == NULL){
		tstrcpy(ErrStr, "EucToSjis: "ERR_ALLOC);
		mem_free(&str);
		return -1;
	}
	EucToSjis(str, ret->v->u.Value);
	ret->v->Type = TYPE_STRING;
	mem_free(&str);
	return 0;
}


/******************************************************************************

	_sfunc_hashtosjis

	nbVɓSJISɕϊ

******************************************************************************/

int SFUNC _sfunc_hashtosjis(EXECINFO *ei, VALUEINFO *param, VALUEINFO *ret, char *ErrStr)
{
	VALUEINFO *vi;
	char *cbuf, *cret, *str;
	int hlen;

	if(param == NULL){
		return -2;
	}

	vi = param;
	hlen = HashToByte(vi, &cbuf);
	str = mem_alloc(hlen + 1);
	CopyMemory(str, cbuf, hlen);
	mem_free(&cbuf);
	*(str + hlen) = '\0';

	cret = mem_alloc(hlen + 1);
	ConvStrCode(str, cret);
	mem_free(&str);

	ret->v->u.Value = cret;
	ret->v->Type = TYPE_STRING;
	return 0;
}


/******************************************************************************

	_sfunc_base64decode

	BASE64̃fR[h

******************************************************************************/

int SFUNC _sfunc_base64decode(EXECINFO *ei, VALUEINFO *param, VALUEINFO *ret, char *ErrStr)
{
	char *str;

	if(param == NULL){
		return -2;
	}

	str = VariableToString(param);

	ret->v->u.Value = mem_alloc(tstrlen(str));
	if(ret->v->u.Value == NULL){
		tstrcpy(ErrStr, "Base64Decode: "ERR_ALLOC);
		mem_free(&str);
		return -1;
	}
	dBase(str, ret->v->u.Value);
	ret->v->Type = TYPE_STRING;
	mem_free(&str);
	return 0;
}


/******************************************************************************

	_sfunc_base64decodehash

	BASE64̃fR[h

******************************************************************************/

int SFUNC _sfunc_base64decodehash(EXECINFO *ei, VALUEINFO *param, VALUEINFO *ret, char *ErrStr)
{
	VALUEINFO *vi;
	char *r;
	char *cret, *rf;

	if(param == NULL){
		return -2;
	}

	//ϐ
	r = VariableToString(param);
	if((param = param->next) == NULL){
		mem_free(&r);
		return -2;
	}

	//nbV
	vi = param->v->vi;

	cret = mem_alloc(tstrlen(r));
	if(cret == NULL){
		tstrcpy(ErrStr, "Base64Decode: "ERR_ALLOC);
		mem_free(&r);
		return -1;
	}
	rf = dBase(r, cret);
	ByteToHash(ei, vi, cret, rf - cret);
	mem_free(&cret);
	mem_free(&r);
	return 0;
}


/******************************************************************************

	_sfunc_base64encode

	BASE64ɃGR[h

******************************************************************************/

int SFUNC _sfunc_base64encode(EXECINFO *ei, VALUEINFO *param, VALUEINFO *ret, char *ErrStr)
{
	char *str;

	if(param == NULL){
		return -2;
	}

	str = VariableToString(param);

	ret->v->u.Value = mem_alloc(tstrlen(str) * 2 + 4);
	if(ret->v->u.Value == NULL){
		tstrcpy(ErrStr, "Base64Encode: "ERR_ALLOC);
		mem_free(&str);
		return -1;
	}
	eBase(str, ret->v->u.Value, 0);
	ret->v->Type = TYPE_STRING;
	mem_free(&str);
	return 0;
}


/******************************************************************************

	_sfunc_base64encodehash

	BASE64ɃGR[h

******************************************************************************/

int SFUNC _sfunc_base64encodehash(EXECINFO *ei, VALUEINFO *param, VALUEINFO *ret, char *ErrStr)
{
	VALUEINFO *vi;
	char *cbuf, *cret;
	int hlen;

	if(param == NULL){
		return -2;
	}

	vi = param;
	hlen = HashToByte(vi, &cbuf);

	cret = mem_alloc(hlen * 2 + 4);
	eBase(cbuf, cret, hlen);
	mem_free(&cbuf);

	ret->v->u.Value = cret;
	ret->v->Type = TYPE_STRING;
	return 0;
}
/* End of source */
