/**************************************************************************

	ns

	func_file.c

	Copyright (C) 1996-2002 by Tomoaki Nakashima. All rights reserved.
		http://www.nakka.com/
		nakka@nakka.com

**************************************************************************/

/**************************************************************************
	Include Files
**************************************************************************/

#include <windows.h>
#include <tchar.h>

#ifdef _WIN32_WCE_LAGENDA
#include <COShellAPI.h>
#endif

#include "script.h"
#include "script_memory.h"
#include "script_string.h"
#include "script_utility.h"


/**************************************************************************
	Define
**************************************************************************/

/**************************************************************************
	Global Variables
**************************************************************************/

typedef struct _FINDINFO {
	HANDLE hFindFile;
	WIN32_FIND_DATA FindData;
} FINDINFO;


/**************************************************************************
	Local Function Prototypes
**************************************************************************/

#ifdef _WIN32_WCE
static TCHAR *CurPath = NULL;
#endif


/******************************************************************************

	GetFullPath

	t@C̃tpX̎擾

******************************************************************************/

#ifdef _WIN32_WCE
static TCHAR *GetFullPath(TCHAR *fname)
{
	TCHAR *ret;

	if(*fname == TEXT('\\') || *fname == TEXT('/')){
		ret = mem_alloc(sizeof(TCHAR) * (lstrlen(fname) + 1));
		lstrcpy(ret, fname);
		if(*(ret + lstrlen(ret) - 1) == TEXT('\\') ||
			*(ret + lstrlen(ret) - 1) == TEXT('/')){
			*(ret + lstrlen(ret) - 1) = TEXT('\0');
		}
		return ret;
	}
	ret = mem_alloc(sizeof(TCHAR) * (lstrlen(CurPath) + lstrlen(fname) + 2));
	wsprintf(ret, TEXT("%s\\%s"), CurPath, fname);
	if(*(ret + lstrlen(ret) - 1) == TEXT('\\') ||
		*(ret + lstrlen(ret) - 1) == TEXT('/')){
		*(ret + lstrlen(ret) - 1) = TEXT('\0');
	}
	return ret;
}
#endif


/******************************************************************************

	ConvPath

	pX̍œK

******************************************************************************/

#ifdef _WIN32_WCE
static TCHAR *ConvPath(TCHAR *path)
{
	TCHAR *p = path, *r;

	while(*p != TEXT('\0')){
		if(*p == TEXT('.')){
			return p;
		}
		for(p = path; *p != TEXT('\0') && *p != TEXT('\\') && *p != TEXT('/'); p++);
		if(*p == TEXT('\\') || *p == TEXT('/')){
			r = ConvPath(p + 1);
			if(*r == TEXT('\0')){
				return r;
			}
			lstrcpy(p, r + 1);
		}
	}
	return p;
}
#endif


/******************************************************************************

	SetCurrentDirectory

	JgfBNg̐ݒ

******************************************************************************/

#ifdef _WIN32_WCE
BOOL SetCurrentDirectory(TCHAR *path)
{
	WIN32_FIND_DATA FindData;
	HANDLE hFindFile;
	TCHAR *buf;
	TCHAR *r;

	if(CurPath == NULL){
		ScriptWait();
		CurPath = mem_alloc(sizeof(TCHAR));
		*CurPath = TEXT('\0');
		ScriptUnWait();
	}
	if(*path == TEXT('\\') || *path == TEXT('/')){
		buf = mem_alloc(sizeof(TCHAR) * (lstrlen(path) + 1));
		lstrcpy(buf, path);
	}else{
		buf = mem_alloc(sizeof(TCHAR) * (lstrlen(CurPath) + lstrlen(path) + 2));
		wsprintf(buf, TEXT("%s\\%s"), CurPath, path);
	}

	r = ConvPath(buf);
	if(*r != TEXT('\0')){
		mem_free(&buf);
		return FALSE;
	}
	if(*(buf + lstrlen(buf) - 1) == TEXT('\\') ||
		*(buf + lstrlen(buf) - 1) == TEXT('/')){
		*(buf + lstrlen(buf) - 1) = TEXT('\0');
	}

	if(*buf != TEXT('\0')){
		hFindFile = FindFirstFile(buf, &FindData);
		if(hFindFile == INVALID_HANDLE_VALUE){
			mem_free(&buf);
			return FALSE;
		}
		FindClose(hFindFile);

		if((FindData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) == 0){
			mem_free(&buf);
			return FALSE;
		}
	}

	ScriptWait();

	mem_free(&CurPath);
	CurPath = mem_alloc(sizeof(TCHAR) * (lstrlen(buf) + 1));
	lstrcpy(CurPath, buf);
	mem_free(&buf);

	ScriptUnWait();
	return TRUE;
}
#endif


/******************************************************************************

	_sfunc_createdirectory

	fBNg̍쐬

******************************************************************************/

int SFUNC _sfunc_createdirectory(EXECINFO *ei, VALUEINFO *param, VALUEINFO *ret, char *ErrStr)
{
#ifdef _WIN32_WCE
	TCHAR *str;
	TCHAR *p;

	if(param == NULL){
		return -2;
	}
	p = VariableToTString(param);
	str = GetFullPath(p);
	mem_free(&p);

	ret->v->u.iValue = (CreateDirectory(str, NULL) == TRUE) ? 0 : -1;
	ret->v->Type = TYPE_INTEGER;
	mem_free(&str);
#else
	TCHAR *str;

	if(param == NULL){
		return -2;
	}
	str = VariableToTString(param);
	ret->v->u.iValue = (CreateDirectory(str, NULL) == TRUE) ? 0 : -1;
	ret->v->Type = TYPE_INTEGER;
	mem_free(&str);
#endif
	return 0;
}


/******************************************************************************

	_sfunc_removedirectory

	fBNg̍폜

******************************************************************************/

int SFUNC _sfunc_removedirectory(EXECINFO *ei, VALUEINFO *param, VALUEINFO *ret, char *ErrStr)
{
#ifdef _WIN32_WCE
	TCHAR *str;
	TCHAR *p;

	if(param == NULL){
		return -2;
	}
	p = VariableToTString(param);
	str = GetFullPath(p);
	mem_free(&p);

	ret->v->u.iValue = (RemoveDirectory(str) == TRUE) ? 0 : -1;
	ret->v->Type = TYPE_INTEGER;
	mem_free(&str);
#else
	TCHAR *str;

	if(param == NULL){
		return -2;
	}
	str = VariableToTString(param);
	ret->v->u.iValue = (RemoveDirectory(str) == TRUE) ? 0 : -1;
	ret->v->Type = TYPE_INTEGER;
	mem_free(&str);
#endif
	return 0;
}


/******************************************************************************

	_sfunc_getcurrentdirectory

	JgfBNg̎擾

******************************************************************************/

int SFUNC _sfunc_getcurrentdirectory(EXECINFO *ei, VALUEINFO *param, VALUEINFO *ret, char *ErrStr)
{
#ifdef _WIN32_WCE
	ret->v->u.Value = tchar2char(CurPath);
#else
	TCHAR str[BUF_SIZE];

	GetCurrentDirectory(BUF_SIZE - 1, str);
#ifdef UNICODE
	ret->v->u.Value = tchar2char(str);
#else
	ret->v->u.Value = alloc_copy(str);
#endif
#endif
	ret->v->Type = TYPE_STRING;
	return 0;
}


/******************************************************************************

	_sfunc_setcurrentdirectory

	JgfBNg̈ړ

******************************************************************************/

int SFUNC _sfunc_setcurrentdirectory(EXECINFO *ei, VALUEINFO *param, VALUEINFO *ret, char *ErrStr)
{
	TCHAR *str;
	BOOL rc;

	if(param == NULL){
		return -2;
	}
	str = VariableToTString(param);

	rc = SetCurrentDirectory(str);

	ret->v->u.iValue = (rc == TRUE) ? 0 : -1;
	ret->v->Type = TYPE_INTEGER;

	mem_free(&str);
	return 0;
}


/******************************************************************************

	_sfunc_copyfile

	t@C̃Rs[

******************************************************************************/

int SFUNC _sfunc_copyfile(EXECINFO *ei, VALUEINFO *param, VALUEINFO *ret, char *ErrStr)
{
#ifdef _WIN32_WCE
	TCHAR *from, *to;
	TCHAR *p;

	if(param == NULL){
		return -2;
	}
	p = VariableToTString(param);
	from = GetFullPath(p);
	mem_free(&p);
	if((param = param->next) == NULL){
		return -2;
	}
	p = VariableToTString(param);
	to = GetFullPath(p);
	mem_free(&p);

	ret->v->u.iValue = (CopyFile(from, to, FALSE) == TRUE) ? 0 : -1;
	ret->v->Type = TYPE_INTEGER;

	mem_free(&from);
	mem_free(&to);
#else
	TCHAR *from, *to;

	if(param == NULL){
		return -2;
	}
	from = VariableToTString(param);
	if((param = param->next) == NULL){
		return -2;
	}
	to = VariableToTString(param);

	ret->v->u.iValue = (CopyFile(from, to, FALSE) == TRUE) ? 0 : -1;
	ret->v->Type = TYPE_INTEGER;

	mem_free(&from);
	mem_free(&to);
#endif
	return 0;
}


/******************************************************************************

	_sfunc_movefile

	t@C̈ړ

******************************************************************************/

int SFUNC _sfunc_movefile(EXECINFO *ei, VALUEINFO *param, VALUEINFO *ret, char *ErrStr)
{
#ifdef _WIN32_WCE
	TCHAR *from, *to;
	TCHAR *p;

	if(param == NULL){
		return -2;
	}
	p = VariableToTString(param);
	from = GetFullPath(p);
	mem_free(&p);
	if((param = param->next) == NULL){
		return -2;
	}
	p = VariableToTString(param);
	to = GetFullPath(p);
	mem_free(&p);

	ret->v->u.iValue = (MoveFile(from, to) == TRUE) ? 0 : -1;
	ret->v->Type = TYPE_INTEGER;

	mem_free(&from);
	mem_free(&to);
#else
	TCHAR *from, *to;

	if(param == NULL){
		return -2;
	}
	from = VariableToTString(param);
	if((param = param->next) == NULL){
		return -2;
	}
	to = VariableToTString(param);

	ret->v->u.iValue = (MoveFile(from, to) == TRUE) ? 0 : -1;
	ret->v->Type = TYPE_INTEGER;

	mem_free(&from);
	mem_free(&to);
#endif
	return 0;
}


/******************************************************************************

	_sfunc_deletefile

	t@C̍폜

******************************************************************************/

int SFUNC _sfunc_deletefile(EXECINFO *ei, VALUEINFO *param, VALUEINFO *ret, char *ErrStr)
{
#ifdef _WIN32_WCE
	TCHAR *str;
	TCHAR *p;

	if(param == NULL){
		return -2;
	}
	p = VariableToTString(param);
	str = GetFullPath(p);
	mem_free(&p);

	ret->v->u.iValue = (DeleteFile(str) == TRUE) ? 0 : -1;
	ret->v->Type = TYPE_INTEGER;

	mem_free(&str);
#else
	TCHAR *str;

	if(param == NULL){
		return -2;
	}

	str = VariableToTString(param);

	ret->v->u.iValue = (DeleteFile(str) == TRUE) ? 0 : -1;
	ret->v->Type = TYPE_INTEGER;

	mem_free(&str);
#endif
	return 0;
}


/******************************************************************************

	_sfunc_fileopen

	t@CJ

******************************************************************************/

int SFUNC _sfunc_fileopen(EXECINFO *ei, VALUEINFO *param, VALUEINFO *ret, char *ErrStr)
{
	HANDLE hFile;
	char *str;
	TCHAR *fname;
#ifdef _WIN32_WCE
	TCHAR *p;
#endif

	if(param == NULL){
		return -2;
	}

	//^Cv
	str = VariableToString(param);
	if((param = param->next) == NULL){
		return -2;
	}

	//t@C
#ifdef _WIN32_WCE
	p = VariableToTString(param);
	fname = GetFullPath(p);
	mem_free(&p);
#else
	fname = VariableToTString(param);
#endif

	//t@CJ
	if(str_cmp_i(str, "write") == 0){
		hFile = CreateFile(fname, GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ, 0, OPEN_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
	}else if(str_cmp_i(str, "new") == 0){
		hFile = CreateFile(fname, GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ, 0, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
	}else{
		hFile = CreateFile(fname, GENERIC_READ, FILE_SHARE_READ | FILE_SHARE_WRITE, 0, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
	}

	if(hFile == NULL || hFile == (HANDLE)-1){
		hFile = (HANDLE)-1;
	}
	ret->v->u.iValue = (long)hFile;
	ret->v->Type = TYPE_INTEGER;

	mem_free(&str);
	mem_free(&fname);
	return 0;
}


/******************************************************************************

	_sfunc_fileclose

	t@C

******************************************************************************/

int SFUNC _sfunc_fileclose(EXECINFO *ei, VALUEINFO *param, VALUEINFO *ret, char *ErrStr)
{
	HANDLE hFile;

	if(param == NULL){
		return -2;
	}

	hFile = (HANDLE)VariableToInt(param);
	if(hFile <= 0){
		tstrcpy(ErrStr, "FileClose: "ERR_HANDLE);
		return -1;
	}
	CloseHandle(hFile);
	return 0;
}


/******************************************************************************

	_sfunc_getfilesize

	t@CTCY̎擾

******************************************************************************/

int SFUNC _sfunc_getfilesize(EXECINFO *ei, VALUEINFO *param, VALUEINFO *ret, char *ErrStr)
{
	HANDLE hFile;
	DWORD fSizeLow;

	if(param == NULL){
		return -2;
	}

	hFile = (HANDLE)VariableToInt(param);
	if(hFile <= 0){
		tstrcpy(ErrStr, "GetFileSize: "ERR_HANDLE);
		return -1;
	}
	fSizeLow = GetFileSize(hFile, NULL);
	ret->v->u.iValue = fSizeLow;
	ret->v->Type = TYPE_INTEGER;
	return 0;
}


/******************************************************************************

	_sfunc_setfilepointer

	t@C|C^̈ړ

******************************************************************************/

int SFUNC _sfunc_setfilepointer(EXECINFO *ei, VALUEINFO *param, VALUEINFO *ret, char *ErrStr)
{
	HANDLE hFile;
	LONG move;
	DWORD method = 0;
	DWORD move_low;
	LONG move_high = 0;

	if(param == NULL){
		return -2;
	}
	//t@Cnh
	hFile = (HANDLE)VariableToInt(param);
	if(hFile <= 0){
		tstrcpy(ErrStr, "SetFilePointer: "ERR_HANDLE);
		return -1;
	}
	if((param = param->next) == NULL){
		return -2;
	}
	move = VariableToInt(param);
	if((param = param->next) != NULL){
		method = VariableToInt(param);
	}

	move_low = SetFilePointer(hFile, move, &move_high, method);
	if(param != NULL && (param = param->next) != NULL){
		param->v->u.iValue = move_high;
		param->v->Type = TYPE_INTEGER;
	}
	ret->v->u.iValue = move_low;
	ret->v->Type = TYPE_INTEGER;
	return 0;
}


/******************************************************************************

	_sfunc_fileread

	t@C̓eǂݏo

******************************************************************************/

int SFUNC _sfunc_fileread(EXECINFO *ei, VALUEINFO *param, VALUEINFO *ret, char *ErrStr)
{
	HANDLE hFile;
	VALUEINFO *vi;
	VALUE v;
	DWORD fSizeLow, fSizeHigh;
	DWORD retsize;
	char *str;

	if(param == NULL){
		return -2;
	}
	//t@Cnh
	hFile = (HANDLE)VariableToInt(param);
	if(hFile <= 0){
		tstrcpy(ErrStr, "FileRead: "ERR_HANDLE);
		return -1;
	}
	if((param = param->next) == NULL){
		return -2;
	}
	//ϐ
	vi = param;
	if((param = param->next) == NULL){
		fSizeLow = GetFileSize(hFile, &fSizeHigh);
	}else{
		//
		fSizeLow = VariableToInt(param);
	}
	if(fSizeLow < 0){
		ret->v->u.iValue = -1;
		ret->v->Type = TYPE_INTEGER;
		return 0;
	}

	str = mem_calloc(fSizeLow + 1);
	if(str == NULL){
		tstrcpy(ErrStr, "FileRead: "ERR_ALLOC);
		return -1;
	}
	if(ReadFile(hFile, str, fSizeLow, &retsize, NULL) == FALSE){
		LocalFree(str);
		ret->v->u.iValue = -1;
		ret->v->Type = TYPE_INTEGER;
		return 0;
	}
	v.u.Value = str;
	v.Type = TYPE_STRING;
	SetValue(ei, vi->v, &v);
	mem_free(&str);

	ret->v->u.iValue = retsize;
	ret->v->Type = TYPE_INTEGER;
	return 0;
}


/******************************************************************************

	_sfunc_filereadhash

	t@C̓eǂݏo

******************************************************************************/

int SFUNC _sfunc_filereadhash(EXECINFO *ei, VALUEINFO *param, VALUEINFO *ret, char *ErrStr)
{
	VALUEINFO *vi;
	HANDLE hFile;
	DWORD fSizeLow, fSizeHigh;
	DWORD retsize;
	char *cBuf;

	if(param == NULL){
		return -2;
	}
	//t@Cnh
	hFile = (HANDLE)VariableToInt(param);
	if(hFile <= 0){
		tstrcpy(ErrStr, "FileRead: "ERR_HANDLE);
		return -1;
	}
	if((param = param->next) == NULL){
		return -2;
	}
	//ϐ
	vi = param->v->vi;
	if((param = param->next) == NULL){
		fSizeLow = GetFileSize(hFile, &fSizeHigh);
	}else{
		//
		fSizeLow = VariableToInt(param);
	}
	if(fSizeLow < 0){
		ret->v->u.iValue = -1;
		ret->v->Type = TYPE_INTEGER;
		return 0;
	}

	cBuf = mem_alloc(fSizeLow + 1);
	if(cBuf == NULL){
		tstrcpy(ErrStr, "FileReadHash: "ERR_ALLOC);
		return -1;
	}
	if(ReadFile(hFile, cBuf, fSizeLow, &retsize, NULL) == FALSE){
		LocalFree(cBuf);
		ret->v->u.iValue = -1;
		ret->v->Type = TYPE_INTEGER;
		return 0;
	}
	ByteToHash(ei, vi, cBuf, retsize);
	mem_free(&cBuf);

	ret->v->u.iValue = retsize;
	ret->v->Type = TYPE_INTEGER;
	return 0;
}


/******************************************************************************

	_sfunc_filewrite

	t@Cɕo

******************************************************************************/

int SFUNC _sfunc_filewrite(EXECINFO *ei, VALUEINFO *param, VALUEINFO *ret, char *ErrStr)
{
	HANDLE hFile;
	DWORD retsize;
	char *p = NULL, *r;
	int len;

	if(param == NULL){
		return -2;
	}
	//t@Cnh
	hFile = (HANDLE)VariableToInt(param);
	if(hFile <= 0){
		tstrcpy(ErrStr, "FileRead: "ERR_HANDLE);
		return -1;
	}
	if((param = param->next) == NULL){
		return -2;
	}
	//ݕ
	if(param->v->Type == TYPE_INTEGER){
		p = r = VariableToString(param);
	}else{
		r = param->v->u.Value;
	}
	if((param = param->next) == NULL){
		len = -1;
	}else{
		//
		len = VariableToInt(param);
	}

	if(len <= 0){
		len = tstrlen(r);
	}
	if(WriteFile(hFile, r, len, &retsize, NULL) == FALSE){
		ret->v->u.iValue = -1;
		ret->v->Type = TYPE_INTEGER;
		mem_free(&p);
		return 0;
	}
	mem_free(&p);
	return 0;
}


/******************************************************************************

	_sfunc_filewritehash

	t@Cɕo

******************************************************************************/

int SFUNC _sfunc_filewritehash(EXECINFO *ei, VALUEINFO *param, VALUEINFO *ret, char *ErrStr)
{
	VALUEINFO *vi;
	HANDLE hFile;
	DWORD retsize;
	int len, hlen;
	char *cBuf;

	if(param == NULL){
		return -2;
	}

	//t@Cnh
	hFile = (HANDLE)VariableToInt(param);
	if(hFile <= 0){
		tstrcpy(ErrStr, "FileRead: "ERR_HANDLE);
		return -1;
	}
	if((param = param->next) == NULL){
		return -2;
	}

	//ϐ
	vi = param;
	if((param = param->next) == NULL){
		len = -1;
	}else{
		//
		len = VariableToInt(param);
	}

	hlen = HashToByte(vi, &cBuf);
	if(len > hlen || len <= 0){
		len = hlen;
	}
	if(WriteFile(hFile, cBuf, len, &retsize, NULL) == FALSE){
		mem_free(&cBuf);
		ret->v->u.iValue = -1;
		ret->v->Type = TYPE_INTEGER;
		return 0;
	}
	mem_free(&cBuf);
	return 0;
}


/******************************************************************************

	_sfunc_filefindinit

	t@C̏

******************************************************************************/

int SFUNC _sfunc_filefindinit(EXECINFO *ei, VALUEINFO *param, VALUEINFO *ret, char *ErrStr)
{
	FINDINFO *FindInfo;
	TCHAR *str;
#ifdef _WIN32_WCE
	TCHAR *p;
#endif

	if(param == NULL){
		return -2;
	}
#ifdef _WIN32_WCE
	p = VariableToTString(param);
	str = GetFullPath(p);
	mem_free(&p);
#else
	str = VariableToTString(param);
#endif

	FindInfo = mem_alloc(sizeof(FINDINFO));
	if(FindInfo == NULL){
		mem_free(&str);
		return 0;
	}

	if((FindInfo->hFindFile = FindFirstFile(str, &FindInfo->FindData)) == INVALID_HANDLE_VALUE){
		mem_free(&FindInfo);
	}
	ret->v->u.iValue = (long)FindInfo;
	ret->v->Type = TYPE_INTEGER;

	mem_free(&str);
	return 0;
}


/******************************************************************************

	_sfunc_filefindnext

	̃t@C

******************************************************************************/

int SFUNC _sfunc_filefindnext(EXECINFO *ei, VALUEINFO *param, VALUEINFO *ret, char *ErrStr)
{
	FINDINFO *FindInfo;

	if(param == NULL){
		return -2;
	}
	FindInfo = (FINDINFO *)VariableToInt(param);
	if(FindInfo == NULL){
		tstrcpy(ErrStr, "FileFindNext: "ERR_HANDLE);
		return -1;
	}

	ret->v->u.iValue = (FindNextFile(FindInfo->hFindFile, &FindInfo->FindData) == TRUE) ? 1 : 0;
	ret->v->Type = TYPE_INTEGER;
	return 0;
}


/******************************************************************************

	_sfunc_filefindclose

	t@C̏I

******************************************************************************/

int SFUNC _sfunc_filefindclose(EXECINFO *ei, VALUEINFO *param, VALUEINFO *ret, char *ErrStr)
{
	FINDINFO *FindInfo;

	if(param == NULL){
		return -2;
	}
	FindInfo = (FINDINFO *)VariableToInt(param);
	if(FindInfo == NULL){
		tstrcpy(ErrStr, "FileFindClose: "ERR_HANDLE);
		return -1;
	}
	FindClose(FindInfo->hFindFile);
	mem_free(&FindInfo);
	return 0;
}


/******************************************************************************

	_sfunc_filefindattributes

	t@C̑擾

******************************************************************************/

int SFUNC _sfunc_filefindattributes(EXECINFO *ei, VALUEINFO *param, VALUEINFO *ret, char *ErrStr)
{
	FINDINFO *FindInfo;

	if(param == NULL){
		return -2;
	}
	FindInfo = (FINDINFO *)VariableToInt(param);
	if(FindInfo == NULL){
		tstrcpy(ErrStr, "FileFindAttributes: "ERR_HANDLE);
		return -1;
	}
	ret->v->u.iValue = FindInfo->FindData.dwFileAttributes;
	ret->v->Type = TYPE_INTEGER;
	return 0;
}


/******************************************************************************

	_sfunc_filefindname

	t@C̃t@C擾

******************************************************************************/

int SFUNC _sfunc_filefindname(EXECINFO *ei, VALUEINFO *param, VALUEINFO *ret, char *ErrStr)
{
	FINDINFO *FindInfo;

	if(param == NULL){
		return -2;
	}
	FindInfo = (FINDINFO *)VariableToInt(param);
	if(FindInfo == NULL){
		tstrcpy(ErrStr, "FileFindName: "ERR_HANDLE);
		return -1;
	}
#ifdef UNICODE
	ret->v->u.Value = tchar2char(FindInfo->FindData.cFileName);
#else
	ret->v->u.Value = alloc_copy(FindInfo->FindData.cFileName);
#endif
	ret->v->Type = TYPE_STRING;
	return 0;
}


/******************************************************************************

	_sfunc_filefindsize

	t@C̃TCY擾

******************************************************************************/

int SFUNC _sfunc_filefindsize(EXECINFO *ei, VALUEINFO *param, VALUEINFO *ret, char *ErrStr)
{
	FINDINFO *FindInfo;

	if(param == NULL){
		return -2;
	}
	FindInfo = (FINDINFO *)VariableToInt(param);
	if(FindInfo == NULL){
		tstrcpy(ErrStr, "FileFindSize: "ERR_HANDLE);
		return -1;
	}
	ret->v->u.iValue = FindInfo->FindData.nFileSizeLow;
	ret->v->Type = TYPE_INTEGER;
	return 0;
}


/******************************************************************************

	_sfunc_filefindtime

	t@C̓擾

******************************************************************************/

int SFUNC _sfunc_filefindtime(EXECINFO *ei, VALUEINFO *param, VALUEINFO *ret, char *ErrStr)
{
	FINDINFO *FindInfo;
	SYSTEMTIME st;
	int type = 0;

	if(param == NULL){
		return -2;
	}
	FindInfo = (FINDINFO *)VariableToInt(param);
	if(FindInfo == NULL){
		tstrcpy(ErrStr, "FileFindTime: "ERR_HANDLE);
		return -1;
	}
	if((param = param->next) != NULL){
		type = VariableToInt(param);
	}
	switch(type)
	{
	case 0:
	default:
		FileTimeToSystemTime(&FindInfo->FindData.ftLastWriteTime, &st);
		break;
	case 1:
		FileTimeToSystemTime(&FindInfo->FindData.ftLastAccessTime, &st);
		break;
	case 2:
		FileTimeToSystemTime(&FindInfo->FindData.ftCreationTime, &st);
		break;
	}
	SystemtimeToVariable(&st, ret);
	return 0;
}


/******************************************************************************

	_sfunc_shell

	wt@Cs

******************************************************************************/

int SFUNC _sfunc_shell(EXECINFO *ei, VALUEINFO *param, VALUEINFO *ret, char *ErrStr)
{
#ifndef _WIN32_WCE_LAGENDA
	SHELLEXECUTEINFO sei;
#endif
	TCHAR *FileName;
	TCHAR *CommandLine = NULL;
	TCHAR *WorkDir = NULL;
	int nShow = SW_SHOWNORMAL;
	BOOL rc;
#ifdef _WIN32_WCE
	TCHAR *p;
#endif

	if(param == NULL){
		return -2;
	}

	//擾
#ifdef _WIN32_WCE
	p = VariableToTString(param);
	FileName = GetFullPath(p);
	mem_free(&p);
#else
	FileName = VariableToTString(param);
#endif
	if((param = param->next) != NULL){
		CommandLine = VariableToTString(param);
		if(*CommandLine == TEXT('\0')){
			mem_free(&CommandLine);
		}
		if((param = param->next) != NULL){
			WorkDir = VariableToTString(param);
			if(*WorkDir == TEXT('\0')){
				mem_free(&WorkDir);
			}
			if((param = param->next) != NULL){
				nShow = VariableToInt(param);
			}
		}
	}

#ifndef _WIN32_WCE_LAGENDA
	ZeroMemory(&sei, sizeof(SHELLEXECUTEINFO));
	sei.cbSize = sizeof(SHELLEXECUTEINFO);
	sei.fMask = SEE_MASK_FLAG_NO_UI;
	sei.lpFile = FileName;
	sei.lpParameters = CommandLine;
#ifdef _WIN32_WCE
	sei.lpDirectory = (WorkDir == NULL) ? CurPath : WorkDir;
#else
	sei.lpDirectory = WorkDir;
#endif
	sei.hInstApp = GetModuleHandle(NULL);
	sei.nShow = nShow;

	//N
	rc = ShellExecuteEx(&sei);
#else
	rc = CoshExecute(NULL, FileName, CommandLine);
#endif

	ret->v->u.iValue = (rc == TRUE) ? 0 : -1;
	ret->v->Type = TYPE_INTEGER;

	mem_free(&FileName);
	mem_free(&CommandLine);
	mem_free(&WorkDir);
	return 0;
}


/******************************************************************************

	_sfunc_execscript

	XNvgs

******************************************************************************/

int SFUNC _sfunc_execscript(EXECINFO *ei, VALUEINFO *param, VALUEINFO *ret, char *ErrStr)
{
	SCRIPTINFO *ScriptInfo;
	EXECINFO cei;
	VALUEINFO *vi;
	VALUEINFO *arg = NULL;
	char *file, *path, *tmp;

	if(CheckEndEvent() == TRUE){
		return -1;
	}

	if(param == NULL){
		return -2;
	}
	tmp = VariableToString(param);
	path = alloc_join(tmp, "\\");
	mem_free(&tmp);
	if((param = param->next) == NULL){
		mem_free(&path);
		return -2;
	}
	file = VariableToString(param);
	if((param = param->next) != NULL){
		arg = param->vi;
	}

	ScriptInfo = mem_alloc(sizeof(SCRIPTINFO));
	if(ScriptInfo == NULL){
		mem_free(&file);
		mem_free(&path);
		return -1;
	}
	InitializeScriptInfo(ScriptInfo, FALSE);
	ScriptInfo->param1 = ei->sci->param1;
	ScriptInfo->param2 = ei->sci->param2;

	ReadScriptFile(ScriptInfo, path, file);
	ExecGlobalArea(ScriptInfo);

	ZeroMemory(&cei, sizeof(EXECINFO));
	cei.sci = ScriptInfo->next;
	vi = ExecFunction(&cei, NULL, "main", arg, FALSE);
	if((int)vi == -1){
		ret->v->u.iValue = -1;
		ret->v->Type = TYPE_INTEGER;
	}else{
		SetValue(ei, ret->v, vi->v);
	}
	FreeValue(vi);

	FreeScriptInfo(ScriptInfo);
	mem_free(&file);
	mem_free(&path);
	return 0;
}
/* End of source */
