/**************************************************************************

	ns

	script_regex.c

	Copyright (C) 1996-2002 by Tomoaki Nakashima. All rights reserved.
		http://www.nakka.com/
		nakka@nakka.com

**************************************************************************/

/**************************************************************************
	Include Files
**************************************************************************/

#include <windows.h>

#include "script.h"

#include "script_regex.h"
#include "script_string.h"
#include "script_memory.h"


/**************************************************************************
	Define
**************************************************************************/

#define ToLower(c)				((c >= 'A' && c <= 'Z') ? (c - 'A' + 'a') : c)

#define CHAR_SPACE_SIZE			256
#define EPSILON_TRANS			-1		// ÑJ

#define CHAR_DOT				"\x01-\xff"


/**************************************************************************
	Global Variables
**************************************************************************/

//Tree
typedef enum {
	TOKEN_CHAR,						// ʏ̕
	TOKEN_GROUP_OPEN,				// K\̃O[v̊Jn
	TOKEN_GROUP_CLOSE,				// K\̃O[v̏I
	TOKEN_INTERVAL_ZERO_MORE_NG,	// Cӂ̐(0܂)̕ (non-greedy)
	TOKEN_INTERVAL_ONE_MORE_NG,		// 1ȏ̕ (non-greedy)
	TOKEN_INTERVAL_ZERO_ONE_NG,		// 01̕ (non-greedy)
	TOKEN_INTERVAL_CLOSE_NG,		// JԂ̎wI (non-greedy)
	TOKEN_INTERVAL_ZERO_MORE,		// Cӂ̐(0܂)̕
	TOKEN_INTERVAL_ONE_MORE,		// 1ȏ̕
	TOKEN_INTERVAL_ZERO_ONE,		// 01̕
	TOKEN_INTERVAL_OPEN,			// JԂ̎wJn
	TOKEN_INTERVAL_CLOSE,			// JԂ̎wI
	TOKEN_ANY_CHAR,					// sCӂ1
	TOKEN_LIST_OPEN,				// Xg̎wJn
	TOKEN_LIST_CLOSE,				// Xg̎wI
	TOKEN_LINE_BEGIN,				// s̐擪
	TOKEN_LINE_END,					// s̖
	TOKEN_ALTERNATION,				// I
	TOKEN_META_ESCAPE,				// ^GXP[v
#ifdef REGVARIABLE
	TOKEN_VARIABLE,					// ϐ
#endif
	TOKEN_END,						// g[N̏I
	TOKEN_LIST_NOT,					// Xg̔ےw
	TOKEN_LIST_RANGE,				// Xg͈͎̔w
	TOKEN_LIST_ESCAPE,				// Xg̃^GXP[v
	TOKEN_INTERVAL_RANGE,			// JԂ͈͎̔w
} token_t;

static char *token_chars[] = {
	"c",	"(",	")",	"*?",	"+?",	"??",	"}?",
	"*",	"+",	"?",	"{",	"}",	".",	"[",
	"]",	"^",	"$",	"|",	"\\",
#ifdef REGVARIABLE
	"%",
#endif
	"",	"^",	"-",	"\\",	",",
};

typedef enum {
	OP_CHAR,						// 
	OP_HEAD,						// s
	OP_TAIL,						// s
	OP_CHARCLASS,					// LN^NX
	OP_CONCAT,						// XY
	OP_UNION,						// X|Y
	OP_CLOSURE,						// X*
	OP_CLOSURE_NG,					// X*?
	OP_EMPTY,						// 
	OP_OPEN,						// (
	OP_CLOSE,						// )
#ifdef REGVARIABLE
	OP_VARIABLE,					// ϐ
#endif
} op_t;


typedef struct tree {
	op_t op;
	union {
		char *chr_class;
		int i;
		struct {
			char c1;
			char c2;
		} c;
#ifdef REGVARIABLE
		struct {
			char *st_var;
			char *end_var;
		} v;
#endif
		struct {
			struct tree *left;
			struct tree *right;
		} x;
	} u;
} tree_t;

typedef struct {
	tree_t *tree;
	token_t current_token;
	char token_char1;
	char token_char2;
	char *p;
	char user_char_set[CHAR_SPACE_SIZE];
	int intervals_open, intervals_close;
#ifdef REGVARIABLE
	char *var;
#endif
	int rfcnt;
} rege_parser_t;


//match
#ifdef REGVARIABLE
typedef struct var {
	char *variable;
	char *st;
	char *en;
} var_t;
#endif

typedef struct rmatch {
	rege_nfa_t *nfa;
	rege_refer_t *rf;
#ifdef REGVARIABLE
	var_t *vt;
#endif
	BOOL icase;
} rege_match_t;


/**************************************************************************
	Local Function Prototypes
**************************************************************************/

//tree
static char *get_esc_string(char *c, char *ret, BOOL expand_flag);
static tree_t *regexp(rege_parser_t *parser);
static void get_token(rege_parser_t *parser);
static tree_t *make_tree_node(rege_parser_t *parser, op_t op, tree_t *left, tree_t *right);
static tree_t *copy_tree(rege_parser_t *parser, tree_t *p);
static tree_t *primary(rege_parser_t *parser);
static tree_t *factor(rege_parser_t *parser);
static tree_t *term(rege_parser_t *parser);
static tree_t *regexp(rege_parser_t *parser);
static rege_parser_t *parse(char *pattern);
static void free_tree(tree_t *p);

//NFA
static nlist_t *add_transition(rege_nfa_t *nfa, int from, int to, int eps, op_t op);
static void gen_nfa(rege_nfa_t *nfa, tree_t *tree, int entry, int way_out);
static int gen_nfa_count(tree_t *tree);
static rege_nfa_t *build_nfa(rege_parser_t *parser);

//match
static BOOL charclass_match(char *p, char *c);
static char *rege_match(rege_match_t *rm, int i, char *c);

#ifdef REGVARIABLE
static void set_variable(var_t *vt, int size, char *var, char *p, BOOL st_flag);
static void reflect_variable(EXECINFO *ei, var_t *vt, int size);
#endif


/******************************************************************************

	get_esc_string

	GXP[v̕ϊ

******************************************************************************/

static char *get_esc_string(char *c, char *ret, BOOL expand_flag)
{
	if(expand_flag == FALSE){
		switch(*(c + 1))
		{
		case '\\':
		case 'w':
		case 'W':
		case 'd':
		case 'D':
		case 's':
		case 'S':
			*ret = *(c++);
			*(ret + 1) = *(c++);
			*(ret + 2) = '\0';
			return c;
		}
	}

	*(ret + 1) = '\0';
	c++;

	switch(*c)
	{
	case 'r':
	case 'R':
		*ret = '\r';
		break;

	case 'n':
	case 'N':
		*ret = '\n';
		break;

	case 't':
	case 'T':
		*ret = '\t';
		break;

	case 'w':
		tstrcpy(ret, "0-9a-zA-Z_");
		break;

	case 'W':
		tstrcpy(ret, "^0-9a-zA-Z_");
		break;

	case 'd':
		tstrcpy(ret, "0-9");
		break;

	case 'D':
		tstrcpy(ret, "^0-9");
		break;

	case 's':
		tstrcpy(ret, "\t\r\n ");
		break;

	case 'S':
		tstrcpy(ret, "^\t\r\n ");
		break;

	case 'x':
	case 'X':
		//Char (Hex)
		c++;
		if(*c != '\0'){
			char tmp[10], *r;

			for(r = c; (r - c) < 2 && ((*r >= '0' && *r <= '9')
				|| (*r >= 'A' && *r <= 'F')
				|| (*r >= 'a' && *r <= 'f')); r++);
			str_cpy_n(tmp, c, r - c + 1);
			*ret = (unsigned char)x2d(tmp);
			c += r - c - 1;
		}
		break;

	default:
		*ret = *c;
		break;
	}
	return (c + 1);
}


/******************************************************************************

	get_token

	g[N̎擾

******************************************************************************/

static void get_token(rege_parser_t *parser)
{
	token_t i;
	int l;
	char *c;

	parser->token_char1 = '\0';
	parser->token_char2 = '\0';
	parser->current_token = TOKEN_CHAR;

	for(i = TOKEN_CHAR + 1; i < TOKEN_END; i++){
		if(*parser->p == '\0'){
			parser->current_token = TOKEN_END;
			return;
		}
		if((l = tstrlen(*(token_chars + i))) &&
			str_cmp_n(parser->p, *(token_chars + i), l) == 0){
			parser->current_token = i;
			break;
		}
	}

	c = parser->p;
	parser->p += tstrlen(*(token_chars + parser->current_token));

	switch(parser->current_token)
	{
	case TOKEN_CHAR:
		parser->token_char1 = *c;
		if(IsDBCSLeadByte((BYTE)*c)){
			parser->token_char2 = *parser->p++;
		}
		break;

	case TOKEN_ANY_CHAR:
		parser->current_token = TOKEN_LIST_CLOSE;
		tstrcpy(parser->user_char_set, CHAR_DOT);
		break;

	case TOKEN_LINE_BEGIN:
		parser->token_char1 = *c;
		break;

	case TOKEN_LINE_END:
		parser->token_char1 = *c;
		break;

#ifdef REGVARIABLE
	case TOKEN_VARIABLE:
		if(*parser->p == **(token_chars + TOKEN_VARIABLE)){
			parser->current_token = TOKEN_CHAR;
			parser->token_char1 = *parser->p;
		}else{
			parser->var = c;
			for(parser->p; *parser->p != **(token_chars + TOKEN_VARIABLE) &&
				*parser->p != '\0'; *parser->p++);
		}
		if(*parser->p != **(token_chars + TOKEN_VARIABLE)){
			parser->current_token = TOKEN_END;
		}else{
			parser->p++;
		}
		break;
#endif

	case TOKEN_INTERVAL_OPEN:
		if(*parser->p != '\0'){
			int *i = &parser->intervals_open;

			parser->intervals_open = 0;
			parser->intervals_close = 0;
			c = parser->p++;
			while(*c != '\0'){
				if(*c == **(token_chars + TOKEN_INTERVAL_CLOSE)){
					if(i == &parser->intervals_open){
						parser->intervals_close = parser->intervals_open;
					}
					parser->current_token = TOKEN_INTERVAL_CLOSE;
					break;

				}else if(*c == **(token_chars + TOKEN_INTERVAL_RANGE)){
					i = &parser->intervals_close;
					c = parser->p;
					parser->p++;

				}else if(*c >= '0' && *c <= '9'){
					*i = *i * 10 + (*c - '0');
					c = parser->p++;

				}else{
					c = parser->p++;
				}
			}
		}
		if(parser->current_token != TOKEN_INTERVAL_CLOSE){
			parser->current_token = TOKEN_END;
		}else{
			if(*parser->p == '?'){
				parser->current_token = TOKEN_INTERVAL_CLOSE_NG;
				parser->p++;
			}
		}
		break;

	case TOKEN_LIST_OPEN:
		if(*parser->p != '\0'){
			int i = 0;

			c = parser->p++;
			if(*c == **(token_chars + TOKEN_LIST_CLOSE)){
				*(parser->user_char_set + i++) = **(token_chars + TOKEN_LIST_CLOSE);
				c = parser->p;
				parser->p++;
			}

			if(*c == **(token_chars + TOKEN_LIST_NOT) &&
				*parser->p == **(token_chars + TOKEN_LIST_CLOSE)){
				*(parser->user_char_set + i++) = **(token_chars + TOKEN_LIST_NOT);
				c = parser->p++;
				*(parser->user_char_set + i++) = **(token_chars + TOKEN_LIST_CLOSE);
				c = parser->p++;
			}

			while(*c != '\0'){
				if(*c == **(token_chars + TOKEN_LIST_CLOSE)){
					*(parser->user_char_set + i) = '\0';
					parser->current_token = TOKEN_LIST_CLOSE;
					break;
				}else if(*c == **(token_chars + TOKEN_LIST_ESCAPE)){
					char tmp[CHAR_SPACE_SIZE];
					parser->p = get_esc_string(c, tmp, FALSE);
					tstrcpy(parser->user_char_set + i, tmp);
					i += tstrlen(tmp);
				}else{
					if(IsDBCSLeadByte((BYTE)*c) == TRUE){
						*(parser->user_char_set + i++) = *c;
						c = parser->p++;
					}
					*(parser->user_char_set + i++) = *c;
				}
				c = parser->p++;
			}
		}
		if(parser->current_token != TOKEN_LIST_CLOSE) {
			parser->current_token = TOKEN_END;
		}
		break;

	case TOKEN_META_ESCAPE:
		if(*parser->p != '\0'){
			parser->p = get_esc_string(c, parser->user_char_set, TRUE);
			parser->current_token = TOKEN_LIST_CLOSE;

			if(tstrlen(parser->user_char_set) == 1){
				parser->current_token = TOKEN_CHAR;
				parser->token_char1 = *parser->user_char_set;
			}
		}
		break;
	}
}


/******************************************************************************

	make_tree_node

	m[h̍쐬

******************************************************************************/

static tree_t *make_tree_node(rege_parser_t *parser, op_t op, tree_t *left, tree_t *right)
{
	tree_t *p;

	if((p = mem_calloc(sizeof(tree_t))) == NULL){
		return NULL;
	}
	p->op = op;
	p->u.x.left = left;
	p->u.x.right = right;
	return p;
}


/******************************************************************************

	copy_tree

	treẽRs[

******************************************************************************/

static tree_t *copy_tree(rege_parser_t *parser, tree_t *p)
{
	tree_t *x;

	if(p == NULL){
		return NULL;
	}
	switch (p->op)
	{
	case OP_CONCAT:
	case OP_UNION:
		x = make_tree_node(parser, p->op,
			copy_tree(parser, p->u.x.left), copy_tree(parser, p->u.x.right));
		break;

	case OP_CLOSURE:
		x = make_tree_node(parser, p->op,
			copy_tree(parser, p->u.x.left), NULL);
		break;

	case OP_CHAR:
		x = make_tree_node(parser, p->op, NULL, NULL);
		x->u.c.c1 = p->u.c.c1;
		x->u.c.c2 = p->u.c.c2;
		break;

	case OP_CHARCLASS:
		x = make_tree_node(parser, p->op, NULL, NULL);
		x->u.chr_class = alloc_copy(p->u.chr_class);
		break;

#ifdef REGVARIABLE
	case OP_VARIABLE:
		x = make_tree_node(parser, p->op, NULL, NULL);
		x->u.v.st_var = p->u.v.st_var;
		x->u.v.end_var = p->u.v.end_var;
		break;
#endif

	default:
		x = make_tree_node(parser, p->op, NULL, NULL);
		break;
	}
	return x;
}


/******************************************************************************

	primary

	X

******************************************************************************/

static tree_t *primary(rege_parser_t *parser)
{
	tree_t *x = NULL;
	tree_t *o, *c;

	switch(parser->current_token)
	{
	case TOKEN_CHAR:
		// X
		x = make_tree_node(parser, OP_CHAR, NULL, NULL);
		x->u.c.c1 = parser->token_char1;
		x->u.c.c2 = parser->token_char2;
		get_token(parser);
		break;

	case TOKEN_LINE_BEGIN:
		// ^
		x = make_tree_node(parser, OP_HEAD, NULL, NULL);
		x->u.c.c1 = parser->token_char1;
		get_token(parser);
		break;

	case TOKEN_LINE_END:
		// $
		x = make_tree_node(parser, OP_TAIL, NULL, NULL);
		x->u.c.c1 = parser->token_char1;
		get_token(parser);
		break;

	case TOKEN_LIST_CLOSE:
		// [...]
		x = make_tree_node(parser, OP_CHARCLASS, NULL, NULL);
		x->u.chr_class = alloc_copy(parser->user_char_set);
		get_token(parser);
		break;

	case TOKEN_GROUP_OPEN:
		// (...)
		parser->rfcnt++;
		o = make_tree_node(parser, OP_OPEN, NULL, NULL);
		o->u.i = parser->rfcnt;
		c = make_tree_node(parser, OP_CLOSE, NULL, NULL);
		c->u.i = parser->rfcnt;

		get_token(parser);
		x = make_tree_node(parser, OP_CONCAT, o,
			make_tree_node(parser, OP_CONCAT, regexp(parser), c));
		get_token(parser);
		break;
	}
	return x;
}


/******************************************************************************

	factor

	X*

******************************************************************************/

static tree_t *factor(rege_parser_t *parser)
{
	tree_t *p, *r, *x;
	int i;

	x = primary(parser);
	switch(parser->current_token)
	{
	case TOKEN_INTERVAL_ZERO_MORE:
		// X*
		x = make_tree_node(parser, OP_CLOSURE, x, NULL);
		get_token(parser);
		break;

	case TOKEN_INTERVAL_ZERO_MORE_NG:
		// X*?
		x = make_tree_node(parser, OP_CLOSURE_NG, x, NULL);
		get_token(parser);
		break;

	case TOKEN_INTERVAL_ONE_MORE:
		// X+
		p = copy_tree(parser, x);
		x = make_tree_node(parser, OP_CONCAT, x,
			make_tree_node(parser, OP_CLOSURE, p, NULL));
		get_token(parser);
		break;

	case TOKEN_INTERVAL_ONE_MORE_NG:
		// X+?
		p = copy_tree(parser, x);
		x = make_tree_node(parser, OP_CONCAT, x,
			make_tree_node(parser, OP_CLOSURE_NG, p, NULL));
		get_token(parser);
		break;

	case TOKEN_INTERVAL_ZERO_ONE:
		// X?
		x = make_tree_node(parser, OP_UNION,
			make_tree_node(parser, OP_EMPTY, NULL, NULL), x);
		get_token(parser);
		break;

	case TOKEN_INTERVAL_ZERO_ONE_NG:
		// X??
		x = make_tree_node(parser, OP_UNION, x,
			make_tree_node(parser, OP_EMPTY, NULL, NULL));
		get_token(parser);
		break;

	case TOKEN_INTERVAL_CLOSE:
		// X{n,m}
		if(parser->intervals_close == 0){
			r = copy_tree(parser, x);
			p = make_tree_node(parser, OP_CLOSURE, r, NULL);
		}else{
			p = make_tree_node(parser, OP_EMPTY, NULL, NULL);
		}
		for(i = parser->intervals_close; i > parser->intervals_open; i--){
			r = copy_tree(parser, x);
			p = make_tree_node(parser, OP_CONCAT, p,
				make_tree_node(parser, OP_UNION,
				make_tree_node(parser, OP_EMPTY, NULL, NULL), r));
		}
		for(i = parser->intervals_open; i > 0; i--){
			r = copy_tree(parser, x);
			p = make_tree_node(parser, OP_CONCAT, r, p);
		}
		if(parser->intervals_open == 0){
			p = make_tree_node(parser, OP_UNION,
				make_tree_node(parser, OP_EMPTY, NULL, NULL), p);
		}
		free_tree(x);
		x = p;
		get_token(parser);
		break;

	case TOKEN_INTERVAL_CLOSE_NG:
		// X{n,m}?
		if(parser->intervals_close == 0){
			r = copy_tree(parser, x);
			p = make_tree_node(parser, OP_CLOSURE_NG, r, NULL);
		}else{
			p = make_tree_node(parser, OP_EMPTY, NULL, NULL);
		}
		for(i = parser->intervals_close; i > parser->intervals_open; i--){
			r = copy_tree(parser, x);
			p = make_tree_node(parser, OP_CONCAT, p,
				make_tree_node(parser, OP_UNION, r,
				make_tree_node(parser, OP_EMPTY, NULL, NULL)));
		}
		for(i = parser->intervals_open; i > 0; i--){
			r = copy_tree(parser, x);
			p = make_tree_node(parser, OP_CONCAT, r, p);
		}
		if(parser->intervals_open == 0){
			p = make_tree_node(parser, OP_UNION, p,
				make_tree_node(parser, OP_EMPTY, NULL, NULL));
		}
		free_tree(x);
		x = p;
		get_token(parser);
		break;

#ifdef REGVARIABLE
	case TOKEN_VARIABLE:
		// variable
		p = NULL;
		if(x == NULL || x->op != OP_CONCAT){
			if(x != NULL){
				p = x;
			}
			x = make_tree_node(parser, OP_CHARCLASS, NULL, NULL);
			x->u.chr_class = alloc_copy(CHAR_DOT);
			if(*parser->p == **(token_chars + TOKEN_END)){
				// (.*)%variable%
				x = make_tree_node(parser, OP_CLOSURE, x, NULL);
			}else{
				// (.*?)%variable%
				x = make_tree_node(parser, OP_CLOSURE_NG, x, NULL);
			}
		}
		x = make_tree_node(parser, OP_CONCAT,
			make_tree_node(parser, OP_VARIABLE, NULL, NULL), x);
		x->u.x.left->u.v.st_var = parser->var;

		x = make_tree_node(parser, OP_CONCAT, x,
			make_tree_node(parser, OP_VARIABLE, NULL, NULL));
		x->u.x.right->u.v.end_var = parser->var;

		if(p != NULL){
			x = make_tree_node(parser, OP_CONCAT, p, x);
		}
		get_token(parser);
		break;
#endif
	}
	return x;
}


/******************************************************************************

	term

	XY

******************************************************************************/

static tree_t *term(rege_parser_t *parser)
{
	tree_t *x;

	if(parser->current_token == TOKEN_ALTERNATION
		|| parser->current_token == TOKEN_GROUP_CLOSE
		|| parser->current_token == TOKEN_END){
		x = make_tree_node(parser, OP_EMPTY, NULL, NULL);

	}else{
		x =  factor(parser);
		while(parser->current_token != TOKEN_ALTERNATION
			&& parser->current_token != TOKEN_GROUP_CLOSE
			&& parser->current_token != TOKEN_END){
			x = make_tree_node(parser, OP_CONCAT, x, factor(parser));
		}
	}
	return x;
}


/******************************************************************************

	regexp

	X|Y

******************************************************************************/

static tree_t *regexp(rege_parser_t *parser)
{
	tree_t *x;

	x = term(parser);
	while(parser->current_token == TOKEN_ALTERNATION){
		get_token(parser);
		x = make_tree_node(parser, OP_UNION, term(parser), x);
	}
	return x;
}


/******************************************************************************

	parse

	\؂̐

******************************************************************************/

static rege_parser_t *parse(char *pattern)
{
	rege_parser_t *parser;

	if((parser = mem_calloc(sizeof(rege_parser_t))) == NULL){
		return NULL;
	}
	parser->p = pattern;
	get_token(parser);

	parser->tree = regexp(parser);
	return parser;
}


/******************************************************************************

	free_tree

	\؂̉

******************************************************************************/

static void free_tree(tree_t *p)
{
	if(p == NULL){
		return;
	}
	switch (p->op)
	{
	case OP_CONCAT:
	case OP_UNION:
		free_tree(p->u.x.right);
		p->u.x.right = NULL;
	case OP_CLOSURE:
	case OP_CLOSURE_NG:
		free_tree(p->u.x.left);
		p->u.x.left = NULL;
		break;

	case OP_CHARCLASS:
		mem_free(&p->u.chr_class);
		break;

	default:
		break;
	}
	mem_free(&p);
}


/******************************************************************************

	add_transition

	NFA ɏԑJڂǉ

******************************************************************************/

static nlist_t *add_transition(rege_nfa_t *nfa, int from, int to, int eps, op_t op)
{
	nlist_t *p;

	if((p = mem_calloc(sizeof(nlist_t))) == NULL){
		return NULL;
	}
	p->eps = eps;
	p->op = op;
	p->to = to;
	p->next = *(nfa->states + from);
	*(nfa->states + from) = p;
	return p;
}


/******************************************************************************

	gen_nfa

	\ tree ɑ΂ NFA 𐶐

******************************************************************************/

static void gen_nfa(rege_nfa_t *nfa, tree_t *tree, int entry, int way_out)
{
	nlist_t *p;
	int a1, a2;

	if(tree == NULL){
		return;
	}
	switch(tree->op)
	{
	case OP_CHAR:
		p = add_transition(nfa, entry, way_out, 0, tree->op);
		if(p != NULL){
			p->u.c.c1 = tree->u.c.c1;
			p->u.c.c2 = tree->u.c.c2;
		}
		break;

	case OP_HEAD:
		if(entry == nfa->entry) {
			nfa->head_flag = TRUE;
			add_transition(nfa, entry, way_out, EPSILON_TRANS, tree->op);
		}
		break;

	case OP_TAIL:
		if(way_out == nfa->exit){
			nfa->tail_flag = TRUE;
			add_transition(nfa, entry, way_out, EPSILON_TRANS, tree->op);
		}
		break;

	case OP_CHARCLASS:
		p = add_transition(nfa, entry, way_out, 0, tree->op);
		if(p != NULL){
			p->u.chr_class = alloc_copy(tree->u.chr_class);
		}
		break;

#ifdef REGVARIABLE
	case OP_VARIABLE:
		p = add_transition(nfa, entry, way_out, EPSILON_TRANS, tree->op);
		if(p != NULL){
			p->u.v.st_var = tree->u.v.st_var;
			p->u.v.end_var = tree->u.v.end_var;
		}
		break;
#endif

	case OP_OPEN:
	case OP_CLOSE:
		p = add_transition(nfa, entry, way_out, EPSILON_TRANS, tree->op);
		if(p != NULL){
			p->u.i = tree->u.i;
		}
		break;

	case OP_EMPTY:
		add_transition(nfa, entry, way_out, EPSILON_TRANS, tree->op);
		break;

	case OP_UNION:
		gen_nfa(nfa, tree->u.x.left, entry, way_out);
		gen_nfa(nfa, tree->u.x.right, entry, way_out);
		break;

	case OP_CLOSURE:
		a1 = nfa->nstate++;
		a2 = nfa->nstate++;
		add_transition(nfa, entry, a1, EPSILON_TRANS, tree->op);
		add_transition(nfa, a2, a1, EPSILON_TRANS, tree->op);
		add_transition(nfa, a1, way_out, EPSILON_TRANS, tree->op);
		gen_nfa(nfa, tree->u.x.left, a1, a2);
		break;

	case OP_CLOSURE_NG:
		a1 = nfa->nstate++;
		a2 = nfa->nstate++;
		add_transition(nfa, entry, a1, EPSILON_TRANS, tree->op);
		gen_nfa(nfa, tree->u.x.left, a1, a2);
		add_transition(nfa, a2, a1, EPSILON_TRANS, tree->op);
		add_transition(nfa, a1, way_out, EPSILON_TRANS, tree->op);
		break;

	case OP_CONCAT:
		a1 = nfa->nstate++;
		gen_nfa(nfa, tree->u.x.left, entry, a1);
		gen_nfa(nfa, tree->u.x.right, a1, way_out);
		break;

	default:
		break;
	}
}


/******************************************************************************

	gen_nfa_count

	ԃXg̐̎擾

******************************************************************************/

static int gen_nfa_count(tree_t *tree)
{
	int ret = 0;

	if(tree == NULL){
		return 0;
	}
	switch (tree->op)
	{
	case OP_UNION:
		ret = 0;
		ret += gen_nfa_count(tree->u.x.left);
		ret += gen_nfa_count(tree->u.x.right);
		break;

	case OP_CLOSURE:
		ret = 2;
		ret += gen_nfa_count(tree->u.x.left);
		break;

	case OP_CLOSURE_NG:
		ret = 2;
		ret += gen_nfa_count(tree->u.x.left);
		break;

	case OP_CONCAT:
		ret = 1;
		ret += gen_nfa_count(tree->u.x.left);
		ret += gen_nfa_count(tree->u.x.right);
		break;

	default:
		break;
	}
	return ret;
}


/******************************************************************************

	build_nfa

	NFA ̐

******************************************************************************/

static rege_nfa_t *build_nfa(rege_parser_t *parser)
{
	rege_nfa_t *nfa;
	int i;

	//NFA ̊m
	if((nfa = mem_calloc(sizeof(rege_nfa_t))) == NULL){
		return NULL;
	}

	//ԃXg̊m
	i = gen_nfa_count(parser->tree);
	nfa->states = mem_calloc(sizeof(nlist_t *) * (i + 2));

	//NFA ̐
	nfa->nstate = 0;
	nfa->entry = nfa->nstate++;
	nfa->exit = nfa->nstate++;
	nfa->head_flag = FALSE;
	nfa->tail_flag = FALSE;
	nfa->rfcnt = parser->rfcnt + 1;
	gen_nfa(nfa, parser->tree, nfa->entry, nfa->exit);
	return nfa;
}


/******************************************************************************

	free_nfa

	NFA ̉

******************************************************************************/

void free_nfa(rege_nfa_t *nfa)
{
	int i;
	nlist_t *p, *q;

	for(i = 0; i < nfa->nstate; i++){
		if(*(nfa->states + i) != NULL){
			p = *(nfa->states + i);
			while(p != NULL){
				q = p;
				p = p->next;
				if(q->op == OP_CHARCLASS){
					mem_free(&q->u.chr_class);
				}
				mem_free(&q);
			}
		}
	}
	mem_free((void **)&nfa->states);
	mem_free(&nfa);
}


/******************************************************************************

	charclass_match

	LN^NX̃}b`O

******************************************************************************/

static BOOL charclass_match(char *p, char *c)
{
	char tmp[CHAR_SPACE_SIZE];
	BOOL ret = FALSE;
	BOOL neg_flag = FALSE;

	if(*p == **(token_chars + TOKEN_LIST_NOT)){
		neg_flag = TRUE;
		p++;
	}

	while(*p != '\0' && ret == FALSE){
		if(IsDBCSLeadByte((BYTE)*p) == TRUE){
			if(*(p + 2) == **(token_chars + TOKEN_LIST_RANGE) && IsDBCSLeadByte((BYTE)*(p + 3)) == TRUE){
				ret = ((unsigned)*c >= (unsigned)*p &&
					(unsigned)*(c + 1) >= (unsigned)*(p + 1) &&
					(unsigned)*c <= (unsigned)*(p + 3) &&
					(unsigned)*(c + 1) <= (unsigned)*(p + 4)) ? TRUE : FALSE;
				p += 4;
			}else{
				ret = (*p == *c && *(p + 1) == *(c + 1)) ? TRUE : FALSE;
				p++;
			}
			p++;
			continue;
		}
		if(*(p + 1) == **(token_chars + TOKEN_LIST_RANGE) && *(p + 2) != '\0'){
			ret = ((unsigned)*c >= (unsigned)*p &&
				(unsigned)*c <= (unsigned)*(p + 2)) ? TRUE : FALSE;
			p += 2;
		}else{
			if(*p == **(token_chars + TOKEN_LIST_ESCAPE) && *(p + 1) != '\0'){
				p = get_esc_string(p, tmp, TRUE);
				ret = charclass_match(tmp, c);
				continue;
			}else{
				ret = (*p == *c) ? TRUE : FALSE;
			}
		}
		p++;
	}
	return ((neg_flag == FALSE) ? ret : !ret);
}


/******************************************************************************

	rege_match

	NFA ŕ̃}b`O

******************************************************************************/

static char *rege_match(rege_match_t *rm, int i, char *c)
{
	nlist_t *p;
	char *ret;

	if(rm->nfa->exit == i){
		if(rm->nfa->tail_flag == TRUE &&
			*c != '\0' && *c != '\r' && *c != '\n'){
			return NULL;
		}
		//
		return c;
	}
	for(p = *(rm->nfa->states + i); p != NULL; p = p->next){
		if(p->eps == EPSILON_TRANS){
			//ÑJ
			switch(p->op)
			{
			case OP_OPEN:
				if(rm->rf == NULL){
					break;
				}
				(rm->rf + p->u.i)->st = c;
				break;

			case OP_CLOSE:
				if(rm->rf == NULL){
					break;
				}
				(rm->rf + p->u.i)->en = c;
				break;

#ifdef REGVARIABLE
			case OP_VARIABLE:
				if(rm->vt == NULL){
					break;
				}
				//ϐ
				if(p->u.v.st_var != NULL){
					set_variable(rm->vt, rm->nfa->nstate / 4, p->u.v.st_var, c, TRUE);
				}
				if(p->u.v.end_var != NULL){
					set_variable(rm->vt, rm->nfa->nstate / 4, p->u.v.end_var, c, FALSE);
				}
				break;
#endif
			}
			if((ret = rege_match(rm, p->to, c)) != NULL){
				return ret;
			}
			continue;
		}

		// ̔r
		if(*c == '\0'){
			continue;
		}
		if(rm->nfa->tail_flag == TRUE && (*c == '\r' || *c == '\n')){
			continue;
		}
		if(IsDBCSLeadByte((BYTE)*c) == TRUE){
			//2oCgR[h
			if(p->op == OP_CHARCLASS){
				//LN^NX
				if(charclass_match(p->u.chr_class, c) == TRUE &&
					(ret = rege_match(rm, p->to, c + 2)) != NULL){
					return ret;
				}
			}else if(p->u.c.c1 == *c && p->u.c.c2 == *(c + 1) &&
				(ret = rege_match(rm, p->to, c + 2)) != NULL){
				return ret;
			}
			continue;
		}
		if(p->op == OP_CHARCLASS){
			//LN^NX
			if(charclass_match(p->u.chr_class, c) == TRUE &&
				(ret = rege_match(rm, p->to, c + 1)) != NULL){
				return ret;
			}
		}else if(rm->icase == TRUE){
			//ignore case
			if(ToLower(p->u.c.c1) == ToLower(*c) &&
				(ret = rege_match(rm, p->to, c + 1)) != NULL){
				return ret;
			}
		}else if(p->u.c.c1 == *c &&
			(ret = rege_match(rm, p->to, c + 1)) != NULL){
			return ret;
		}
	}
	return NULL;
}


/******************************************************************************

	rege_compile

	K\RpC

******************************************************************************/

rege_nfa_t *rege_compile(char *Ptn)
{
	rege_parser_t *parser;
	rege_nfa_t *nfa;

	//\؂̐
	parser = parse(Ptn);

	//NFA̐
	nfa = build_nfa(parser);
	free_tree(parser->tree);
	mem_free(&parser);
	return nfa;
}


/******************************************************************************

	reg_match

	K\̃}b`O

******************************************************************************/

int reg_match(rege_nfa_t *nfa, char *Str, rege_refer_t **rf, BOOL icase)
{
	rege_match_t rm;
	char *to;
	char *p;

	to = NULL;

	if(nfa == NULL){
		return 0;
	}

	ZeroMemory(&rm, sizeof(rege_match_t));
	rm.nfa = nfa;
	rm.icase = icase;
	rm.rf = mem_alloc(sizeof(rege_refer_t) * nfa->rfcnt);

	p = Str;
	while(*p != '\0'){
		ZeroMemory(rm.rf, sizeof(rege_refer_t) * nfa->rfcnt);
		to = rege_match(&rm, 0, p);
		if(to != NULL){
			break;
		}
		if(nfa->head_flag == TRUE){
			for(; *p != '\0' && *p != '\r' && *p != '\n'; p++);
			for(; *p == '\r' || *p == '\n'; p++);
		}else{
			p++;
		}
	}
	if(to == NULL){
		mem_free(&rm.rf);
		return 0;
	}
	rm.rf->st = p;
	rm.rf->en = to;
	*rf = rm.rf;
	return nfa->rfcnt;
}


/******************************************************************************

	set_variable

	ϐ͈͂̐ݒ

******************************************************************************/

#ifdef REGVARIABLE
static void set_variable(var_t *vt, int size, char *var, char *p, BOOL st_flag)
{
	int i;

	for(i = 0; i < size; i++){
		if((vt + i)->variable == NULL){
			break;
		}
		if((vt + i)->variable == var){
			break;
		}
	}
	if((vt + i)->variable == NULL){
		(vt + i)->variable = var;
	}
	if(st_flag == TRUE){
		(vt + i)->st = p;
	}else{
		(vt + i)->en = p;
	}
}
#endif


/******************************************************************************

	reflect_variable

	ϐ̔f

******************************************************************************/

#ifdef REGVARIABLE
static void reflect_variable(EXECINFO *ei, var_t *vt, int size)
{
	VALUEINFO *vi;
	VALUE vl;
	char *r;
	int i;

	if(ei == NULL){
		return;
	}
	for(i = 0; i < size; i++){
		if((vt + i)->variable == NULL){
			break;
		}
		if((vt + i)->st == NULL || (vt + i)->en == NULL){
			r = alloc_copy("");
		}else{
			r = alloc_copy_n((vt + i)->st, (vt + i)->en - (vt + i)->st + 1);
		}

		vi = VariableFromString(ei, (vt + i)->variable);
		if(vi != NULL){
			vl.u.Value = r;
			vl.Type = TYPE_STRING;
			SetValue(ei, vi->v, &vl);
			FreeValue(vi);
		}
		mem_free(&r);
	}
}
#endif


/******************************************************************************

	reg_match_value

	K\̃}b`OsAϐɊi[

******************************************************************************/

BOOL reg_match_value(EXECINFO *ei, rege_nfa_t *nfa, char *Str)
{
	rege_match_t rm;
	nlist_t *q;
	char *p;
	char *ret = NULL;
	int i;
	BOOL rc;

	if(nfa == NULL){
		return FALSE;
	}

	ZeroMemory(&rm, sizeof(rege_match_t));
	rm.nfa = nfa;
	rm.icase = FALSE;

#ifdef REGVARIABLE
	//ϐ̏
	rm.vt = mem_calloc(sizeof(var_t) * (nfa->nstate / 4));
	for(i = 0; i < nfa->nstate; i++){
		for(q = *(nfa->states + i); q != NULL; q = q->next){
			if(q->op == OP_VARIABLE && q->u.v.st_var != NULL){
				set_variable(rm.vt, nfa->nstate / 4, q->u.v.st_var, NULL, TRUE);
			}
		}
	}
#endif

	p = Str;
	while(*p != '\0'){
		ret = rege_match(&rm, 0, p);
		if(ret != NULL){
			break;
		}
		if(nfa->head_flag == TRUE){
			for(; *p != '\0' && *p != '\r' && *p != '\n'; p++);
			for(; *p == '\r' || *p == '\n'; p++);
		}else{
			p++;
		}
	}
	rc = ((ret != NULL) ? TRUE : FALSE);

#ifdef REGVARIABLE
	if(rc == FALSE){
		//}b`Ȃꍇ͒lNA
		for(i = 0; i < nfa->nstate / 4; i++){
			rm.vt[i].st = NULL;
		}
	}
	reflect_variable(ei, rm.vt, nfa->nstate / 4);
	mem_free(&rm.vt);
#endif
	return rc;
}
/* End of source */
